/* 
 * easter.c   Calculates easter sunday          1999-2-13
 * Copyright (C) 1999 Antonio Luque Estepa  <aluque@esi.us.es>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#if(!defined HAVE_GETOPT_H)
#error Must have <getopt.h>
#endif

#include <stdio.h>
#include "getopt.h"
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

//For debugging purposes only
#define DEBUG

#define SUNDAY 0
#define MONDAY 1
#define TUESDAY 2
#define WEDNESDAY 3
#define THURSDAY 4
#define FRIDAY 5
#define SATURDAY 6

#define isleap(yr) \
(((yr) <= 1582 ) ? !((yr)%4) : \
(!((yr)%4)) && (((yr)%100) || !((yr)%400)))

//Number of julian leap years that are not gregorian leap years since 1582
#define notleapsince1582(yr) \
(((yr) >= 1600 ? (yr)/100 -16: 0) -((yr) >= 2000 ? (yr)/400 -4: 0)) 

//Number of leap years since year 1
#define leapsince1(yr) \
( (yr)/4 - notleapsince1582(yr) )

#define JULIAN5OCT1582 577738
#define JD0 2415020L

void main(int, char **);
void usage(void);
void version(void);
int dayofweek(int,int,int);
int jday(int,int,int);
void juliantodate(int,int,int *,int *);
int yeartoeaster(int *,int *,int);
int daysto(int,int,int);
double juliannr(int,int,int);

char const progname[]="easter  2.1.1  (c) 1999 Antonio Luque <aluque@esi.us.es>\n";
char const usagemesg[]="easter [-rac] [-f] [-hvV] year\n"
"easter calculates easter of given year.\n"
"\t-r Calculates Palm Sunday, instead of easter.\n"
"\t-c Calculates Ash Wednesday.\n"
"\t-a Displays all data (Ash, Lent, Palm and Easter).\n"
"\t-f Number of days resting to Easter (or Palm, or Ash).\n"
"\t-V Verbose mode.\n"
"\t-v Shows version number and exits\n"
"\t-h Shows this text and exits\n"
"If year is not given, use current.\n"
"Note that only one of `-r' `-c' `-a' can be specified.\n"
"All dates are displayed in day-month format.\n";

int verbose;

typedef struct {
    int day;
    int month;
    int year;
} date;

//Table of Paschal full moons depending on the epacta
date paschal[]={
   {0,0}, {12,4}, {11,4}, {10,4}, {9,4}, {8,4}, {7,4}, {6,4}, {5,4}, {4,4},
   {3,4}, {2,4},  {1,4},  {31,3}, {30,3},{29,3},{28,3},{27,3},{26,3},{25,3},
   {24,3},{23,3}, {22,3}, {21,3}, {18,4},{0,0}, {17,4},{16,4},{15,4},{14,4},
   {13,4}
};
//If epacta==25, paschal moon is determined by golden number

//Number of days in each month
int monthday[]=
{ 31,28,31,30,31,30,31,31,30,31,30,31 };
int leapmonthday[]=
{ 31,29,31,30,31,30,31,31,30,31,30,31 };

char *dayname[7]=
{ "Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday" };

void main(int argc, char *argv[])
{
    int c,i;
    char opciones[]="rchvVaf";
    int year;
    date easter; //Date of easter
    date ashdate;
    date palmdate;
    date lentdate[5];  //Dates of every Lent Sunday
    date thisdate;
    int easterjd;  //Easter Julian day
    int palm=0;
    int ash=0;
    int all=0;
    int rest=0;
    struct timeval tv;
    struct tm *ttt;
    char what[16]; //What are we calculating
    
    if(argc<1) { //Wrong number of arguments
	usage();
	exit(0);
    }
    
    //Parse command line options
    while((c=getopt(argc,argv,opciones))!=-1){
	switch(c) {
	  case 'r':
	    //calculate palm sunday
	    if(ash || all) {
		usage();
		exit(0);
	    }
	    palm=1;
	    break;
	  case 'c':
	    //calculate ash wednesday
	    if(palm || all) {
		usage();
		exit(0);
	    }
	    ash=1;
	    break;
	  case 'a':
	    if(palm || ash) {
		usage();
		exit(0);
	    }
	    all=1;
	    break;
	  case 'f':
	    if(all) {
		printf("easter: days resting to what?\n");
		exit(1);
	    }
	    rest=1;
	    break;	    
	  case 'V':
	    verbose=1;
	    break;
	  case 'v':
	    version();
	    exit(1);
	    break;
	  case 'h':
	    usage();
	    exit(1);
	    break;
	  case '?':
	    if(isprint(optopt))
		fprintf(stderr,"%s -h for help\n",argv[0]);
	    else
		fprintf(stderr,"Invalid character: -%c\n%s -h for help\n",optopt,argv[0]);
	    return;
	  default:
	    abort();
	}
    }
    
    if(optind==argc) { //No year was specified
	gettimeofday(&tv,NULL);
	ttt=localtime((time_t *)&tv.tv_sec);
	year=ttt->tm_year+1900;
    }
    else {
	if(argv[optind][0]=='-') {
	    if(!isdigit(argv[optind][1])) {
		printf("easter: Can not understand year.\n");
		exit(0);
	    }
	}
	else if(!isdigit(argv[optind][0])) {
	    printf("easter: Can not understand year.\n");
	    exit(0);	    
	}
	sscanf(argv[optind],"%d",&year);
    }
    if(year==0) {
	fprintf(stderr,"easter: there was no year 0.\n");
	exit(0);
    }
    
    //Calculate all relevant dates
    easterjd=yeartoeaster(&easter.day,&easter.month,year);
    juliantodate(easterjd-7,year,&palmdate.day,&palmdate.month);
    juliantodate(easterjd-46,year,&ashdate.day,&ashdate.month);
    for(i=0;i<5;++i)
	juliantodate(easterjd-7*(i+2),year,&lentdate[4-i].day,&lentdate[4-i].month);

    
    //Display appropriate dates
    if(palm) {
	strcpy(what,"Palm");
	thisdate=palmdate;
    }
    else if(ash) {
	strcpy(what,"Ash");
	thisdate=ashdate;
    }
    else {
	strcpy(what,"Easter");
	thisdate=easter;
    }

    if(rest) {
	int to;
	to=daysto(thisdate.day,thisdate.month,year);
	while(to<0) {
	    year++;
	    //FIXME: make possible this date being another than Easter
	    easterjd=yeartoeaster(&easter.day,&easter.month,year);
	    thisdate=easter;
	    to=daysto(thisdate.day,thisdate.month,year);
	    if(to>=0) fprintf(stderr,"Displaying days resting to %s of %d:\n",
			     "Easter",year);
	}
	printf("%d\n",to);
	return;
    }
    
    if(all)  {
	printf("Calendar for %d:\n",year);
	printf("  Ash      %d-%d\n"
	       "  I   Lent %d-%d\n"
	       "  II  Lent %d-%d\n"
	       "  III Lent %d-%d\n"
	       "  IV  Lent %d-%d\n"
	       "  V   Lent %d-%d\n"
	       "  Palm     %d-%d\n"
	       "  Easter   %d-%d\n",ashdate.day,ashdate.month,
	       lentdate[0].day,lentdate[0].month,lentdate[1].day,lentdate[1].month,
	       lentdate[2].day,lentdate[2].month,lentdate[3].day,lentdate[3].month,
	       lentdate[4].day,lentdate[4].month,
	       palmdate.day,palmdate.month,easter.day,easter.month);
    }
    else {
	if(verbose)
	    printf("%s for year %d:\n",what,year);
	printf("%d %d\n",thisdate.day,thisdate.month);
    }
}

void usage(void)
{
    fprintf(stdout,progname);
    fprintf(stdout,usagemesg);
}

void version(void)
{
    fprintf(stdout,progname);
}

//dayofweek returns the day of the week, given the complete date
int dayofweek(int day,int month,int year)
{
    long julianday; //Julian day total since 1 Jan 1
    int weekday=THURSDAY;
    
    julianday=(year-1)*365+leapsince1(year-1)+jday(day,month,year);
    
    if(julianday < JULIAN5OCT1582) { //Julian calendar
	weekday=((julianday-1 + SATURDAY)%7);
    }

    if(julianday >= JULIAN5OCT1582+9){
	weekday=((julianday-1+SATURDAY - 10)%7);
    }
    
    if(verbose) {
	fprintf(stderr,"%d-%d-%d is %s\n",day,month,year,dayname[weekday]);
    }
    
    return weekday;
}

//Calculates the julian day (day number within the year)
int jday(int day, int month,int year)
{
    int temp=0,i;
    if(isleap(year)){
	for(i=0;i<month-1;++i)
	    temp+=leapmonthday[i];
	temp+=day;
    }
    else {
	for(i=0;i<month-1;++i)
	    temp+=monthday[i];
	temp+=day;
    }
    return temp;
}

//Converts a julian date to a normal date
void juliantodate(int julian,int year, int *day,int *month)
{
    int i;
    if(isleap(year)) {
	for(i=0;i<12;++i){
	    julian-=leapmonthday[i];
	    if(julian<=0) {
		*month=i+1; *day=julian+leapmonthday[i];
		return;
	    }
	}
    }
    else {   
	for(i=0;i<12;++i) {
	    julian-=monthday[i];
	    if(julian<=0) { 
		*month=i+1; *day=julian+monthday[i]; 
		return;
	    }
	}
    }
}

//Calculates easter for a given year
//    day, month: Easter date
//    year: year
// returns: julian day of easter in that year
int yeartoeaster(int *day,int *month,int year)
{
    int cent;  //Century
    date full; //Paschal Full moon
    date easter;
    int gn,ep;  //Golden Number, Epacta
    int julianday,weekday;
    
    //When will 21st century start? 
    //I dont want to dispute about the same again
    //For now, we assume century=two first digits of year + 1
    cent=((year-(year%100))/100)+1;
    if(verbose) {
	fprintf(stderr,"Calculating Easter for year %d\n",year);
	fprintf(stderr,"Year is %d, and is%s leap. Century is %d\n",year,
		isleap(year)?"":" not",cent);
    }
    
    gn=(year%19)+1;
    ep=(11*(gn-1))%30;
    ep=(int)(ep-(3*cent)/4);
    ep=(int)(ep+(8*cent+5)/25);
    ep+=8;
    while(ep<=0) ep+=30;
    while(ep>30) ep-=30;
        
    if(verbose) {
	fprintf(stderr,"Golden Number is %d. Epacta is %d\n",gn,ep);
    }
    
    full=paschal[ep];
    if(ep==25) {
	if(gn>11) { full.day=17; full.month=4; };
	if(gn<=11) { full.day=18; full.month=4; };
    }
    
    if(verbose) {
	fprintf(stderr,"Paschal full moon is %d-%d",full.day,full.month);
    }
    
    //Now we must calculate the sunday following the paschal full moon
    
    julianday=jday(full.day,full.month,year);   
    
    if(verbose) {
	fprintf(stderr,", which is the %dth day of year\n",julianday);
    }
    weekday=dayofweek(full.day,full.month,year);
    
    julianday+=7-weekday;
    
    if(verbose) {
	fprintf(stderr,"Easter is %dth day of year",julianday);
    }
    
    juliantodate(julianday,year,&easter.day,&easter.month); 
    
    if(verbose) {
	fprintf(stderr,", that is %d-%d\n",easter.day,easter.month);
    }
    
    *day=easter.day;
    *month=easter.month;
    return julianday;
}

//Days resting to a date
//   day,month,year: date
// returns: days resting
int daysto(int day,int month,int year)
{
    double jdtoday,jddate;
    date today;
    
    struct timeval tv;
    struct tm *ttt;
    
    gettimeofday(&tv,NULL);
    ttt=localtime((time_t *)&tv.tv_sec);
    today.year=ttt->tm_year+1900;
    today.day=ttt->tm_mday;
    today.month=ttt->tm_mon+1;
    jdtoday=juliannr(today.day,today.month,today.year);
    
    jddate=juliannr(day,month,year);
    
    return (int)(jddate-jdtoday);
}

//Calculates Julian number for given date
//This Julian number has year in consideration (is therefore, the official
//astronomical Julian number). Note the difference with jday()
//   day,month,year: date
// returns; Julian number
//This function uses code based on:
// XEphem (c) 1990-97 Elwood Charles Downey <ecdowney@noao.edu>
double juliannr(int day,int month,int year)
{
    int b,d,m,y;
    double JD;
    long c;
    
    m=month;
    y=(year < 0) ? year+1 : year;
    if(month<3) {
	m+=12;
	y-=1;
    }
    if(year<1582 || (year==1582 && (month<10  || (month==10 && day<15))))
	b=0;
    else {
	int a;
	a=y/100;
	b=2-a+a/4;
    }
    if(y<0)
	c=(long)((365.25*y) - 0.75) - 694025L;
    else
	c=(long)(365.25*y) - 694025L;
    
    d=30.6001*(m+1);
    JD=b+c+d+day;
    
    JD+=JD0;
    
    if(verbose)
	fprintf(stderr,"Julian number for %d-%d-%d is %.0f\n",year,month,day,JD);
    
    return JD;
}
